/*
 * SPDX-License-Identifier: AGPL-3.0-only
 * Copyright (c) Camden Dixie O'Brien
 */

#include "time_manager.h"

#include "console.h"
#include "settings.h"

#include "esp_log.h"
#include "esp_sntp.h"
#include <time.h>

#define TAG "Time"

static void handle_timezone_update(const char *timezone)
{
	setenv("TZ", timezone, 1);
	tzset();
	(void)sntp_restart();
}

static void handle_sntp_server_update(const char *sntp_server)
{
	esp_sntp_setservername(0, sntp_server);
	if (!sntp_restart()) {
		ESP_LOGW(
		    TAG, "Received SNTP server update before SNTP initialization");
	}
}

static void sntp_sync_callback(struct timeval *tv)
{
	const time_t now = tv->tv_sec;
	struct tm timeinfo;
	(void)localtime_r(&now, &timeinfo);
	ESP_LOGI(
	    TAG, "Received SNTP time notification: %02u:%02u", timeinfo.tm_hour,
	    timeinfo.tm_min);
}

static const char *sync_status_description(sntp_sync_status_t status)
{
	switch (status) {
	case SNTP_SYNC_STATUS_RESET:
		return "Reset";
	case SNTP_SYNC_STATUS_COMPLETED:
		return "Completed";
	case SNTP_SYNC_STATUS_IN_PROGRESS:
		return "In progress";
	default:
		return "Invalid";
	}
}

static int command_func(int argc, char **argv)
{
	if (argc == 1) {
		const Time time = get_time();
		printf("%02u:%02u\n", time.hour, time.minute);
		return 0;
	} else if (argc == 2) {
		if (strcmp(argv[1], "sntp-status") == 0) {
			const sntp_sync_status_t status = sntp_get_sync_status();
			printf("%s\n", sync_status_description(status));
			return 0;
		} else {
			Time time;
			if (sscanf(argv[1], "%02u:%02u", &time.hour, &time.minute) < 2
			    || time.hour > 23 || time.minute > 59) {
				printf("Invalid time\n");
				return 1;
			}
			set_time(time);
			return 0;
		}
	} else {
		printf("Invalid number of arguments\n");
		return 1;
	}
}

void time_manager_init(void)
{
	char timezone[SETTINGS_MAX_VALUE_SIZE];
	(void)settings_get_timezone(timezone, SETTINGS_MAX_VALUE_SIZE);
	handle_timezone_update(timezone);
	settings_add_timezone_callback(&handle_timezone_update);

	char sntp_server[SETTINGS_MAX_VALUE_SIZE];
	(void)settings_get_sntp_server(sntp_server, SETTINGS_MAX_VALUE_SIZE);
	esp_sntp_set_sync_mode(SNTP_SYNC_MODE_IMMED);
	esp_sntp_setoperatingmode(ESP_SNTP_OPMODE_POLL);
	esp_sntp_setservername(0, sntp_server);
	sntp_set_time_sync_notification_cb(sntp_sync_callback);
	esp_sntp_init();
	settings_add_sntp_server_callback(&handle_sntp_server_update);

	console_register(
	    "time", "Get / set time and SNTP status",
	    "time OR time <HH:MM> OR time sntp-status", command_func);
}

Time get_time(void)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	struct tm timeinfo;
	(void)localtime_r(&tv.tv_sec, &timeinfo);

	return (Time) { .hour = timeinfo.tm_hour, .minute = timeinfo.tm_min };
}

void set_time(Time time)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	struct tm timeinfo;
	(void)localtime_r(&tv.tv_sec, &timeinfo);

	timeinfo.tm_hour = time.hour;
	timeinfo.tm_min = time.minute;
	tv.tv_sec = mktime(&timeinfo);
	settimeofday(&tv, NULL);
}
